'''Game main module.

Contains the entry point used by the run_game.py script.
'''

from concepts.color import *
from concepts.logger import *
from concepts.geometry import *

from river.building import *
from river.gui import *
from river.scenario import *
from river.terrain import *
from river.industries.farming import *
from river.industries.logging import *
from river.industries.storage import *

from ui.components import *

import os


def main(data_dir):
    Component.set_font(os.path.join(data_dir, 'FreeSans'), 14)
    
    logger   = Logger()
    scenario = TestScenario()

    scenario.start(logger)
    
    size   = Vector(1024, 768)
    window = RiverWindow(size, scenario, logger)
    
    window.do_events()
    
    scenario.stop()
    

class TestScenario(Scenario):
    def setup(self):
        tile     = self.map().tile
        build    = self.city().build
        birth    = self.city().give_birth
        x_size   = self.map_size().x()
        y_size   = self.map_size().y()
        x_center = x_size / 2
        y_center = y_size / 2
        layout   = ['$$$$$=$$$$$$$$$$$$=$=$$$$',
                    '$=$$$$...$$$$=$$$$$$$$$$$',
                    '$$$$...~~~~$$$$$$$$$$$=$$',
                    '$$....~~~~~~$$$$$$=$$$$$$',
                    '=$$$.~~~..~~~$=$$$$$$$$$$',
                    '$$..~~~....~~~~$$$$$$~~=$',
                    '=$=.~~~.....~~~$$$$$~~~~$',
                    '$$=$.~~~..~~~~$$=$~~~~~~~',
                    '=$$..~~~..~~~~~~~~~~...~~',
                    '=....~~~...~~~~~~~~....~~',
                    '....~~~.......~~~.....~~~',
                    '...~~~................~~~',
                    '...~~~...............~~~.',
                    '..~~~...............~~~..',
                    '...~~~.........~~~~~~~...',
                    '..~~~~~.....~~~~~~~~~....',
                    '~~~~~~....~~~~~..........',
                    '~~~~.....~~~.............',
                    '~~........~~~~..~~~~~....',
                    '............~~~~~~~~~~~~.',
                    '~~~~~~~......~~~...~~~~~~',
                    '~~~~~~~~~..........~~~~..',
                    '~~~~~~~~~~.......~~~~....',
                    '......~~~~~~...~~~~......',
                    '.......~~~~~~~~~~........']
        
        self.allow(Farm)
        
        for y in range(len(layout)):
            line = layout[y]
            
            for x in range(len(line)):
                c = line[x]
                t = tile(x, y)
                
                if c == '~':
                    t.become(River)                    
                
                if c == '$':
                    t.become(Forest)
                
                if c == '=':
                    t.become(Forest)
                    t.spawn(Wood(2))
                    
                if c == 'r':
                    t.build(Road)
                
                if c == 'b':
                    t.become(River)                    
                    t.build(Bridge)
                
                if c == 'h':
                    t.build(House)
                
                if c == 'f':
                    t.build(Farm)

    def schedule(self):      
        y_center  = self.map_size().y() / 2
        tile      = self.map().tile(0, y_center)
        immigrant = lambda: self.city().give_birth(tile)
        
        return {  1: [immigrant],
                 20: [immigrant],
                 40: [immigrant]}
    
    def more_people(self, num):
        x_end     = self.map_size().x() - 1
        y_center  = self.map_size().y() / 2
        tile      = self.map().tile(x_end, y_center)
        
        for _ in range(num):
            self.city().give_birth(tile)
        
        self.message('%d immigrants arrived!' % num, Color(192, 192, 255))
        
    def intro(self):
        return [('Twisted River Community - Tutorial', Color(192, 192, 255)),
                ('  Welcome to the Twisted River full of twisted people! This tutorial will show you how to build a nice little town while', None),
                ('  avoiding starving everyone. You start with three people, who are already getting hungry from the journey here! ', None),
                ('Your first task: Build a farm!', Color(255, 255, 0)),
                ('  You can build buildings from the menu to the lower right. A villager will become builder and gather the materials.', Color(128, 128, 128))]

    def script(self):
        number_of_p = lambda x: len(self.city().persons_of_type(x))
        number_of_b = lambda x: len(self.city().buildings_of_type(x))
        two_more    = lambda: self.more_people(2)
        
        return [(lambda: number_of_p('Builder')    ==  1, self.first_farm     ),
                (lambda: number_of_p('Farmer')     ==  2, self.more_wood      ),
                (lambda: number_of_p('Lumberjack') ==  1, self.immigrants     ),
                (lambda: number_of_b(House)        ==  5, self.more_immigrants),
                (lambda: number_of_b(House)        ==  6, two_more            ),
                (lambda: number_of_b(House)        ==  7, two_more            ),
                (lambda: number_of_b(House)        ==  8, two_more            ),
                (lambda: number_of_b(House)        ==  9, two_more            ),
                (lambda: number_of_b(House)        == 10, self.won            )]
        
    def first_farm(self):
        name = self.city().persons_of_type('Builder')[0].name()
        
        self.message('Well done! %s became a builder and is now gathering wood for the farm.' % name)
        self.hint('There are lots of stacks of two wood each in the forest to the north.')
        self.hint('If you click on a tile while not in construction mode, you select a random person on it.')
        
    def more_wood(self):
        other = self.city().persons_of_type('Villager')[0].name()
        names = ' and '.join(map(lambda x: x.name(),
                                 self.city().persons_of_type('Farmer')))
        
        self.allow(Sawmill)

        self.message('The farm is done! Every farm employs up to two people. %s are now farmers and will grow wheat.' % names)
        self.message('%s is still a villager without job and should now construct a sawmill to ensure your wood supply.' % other)
        self.hint('All sawmill tiles need to be inside the forest.')
    
    def immigrants(self):
        name = self.city().persons_of_type('Lumberjack')[0].name()
        
        self.allow(Road)
        self.allow(Bridge)
        self.allow(House)
        
        self.message('And with that, %s has a job as lumberjack! Unfortunatly, nobody is able to be a builder now.' % name)
        self.message('Time for some immigrants! But be aware, those are more mouths to feed!')
        self.message('Houses will attract more immigrants (for this version). Build five of them!')
        self.hint('You can now build infrastructure. Roads and bridges make people run faster on them.')
        
        self.more_people(5)
        
    def more_immigrants(self):
        self.allow(Storage)
        
        self.message('Great! Two more people joined the settlement and brought plans for a storage hall with them.')
        self.message('Every new house will now bring two more immigrants. Try to reach twenty people in your town!')
        self.hint('Storage buildings employ three collectors who will collect all resources and bring them to the building.')
        
        self.more_people(2)

    def won(self):
        self.message('Yay! You got 20 people and won!!!')
        self.hint('That\'s all there is for the moment.')
        self.message('Thank you for playing!', Color(255, 0, 0))
        
        self.more_people(2)
        