'''Movement actions.'''

from concepts.color import *
from concepts.geometry import *
    
from _base import *

import random


class Wait(Action):
    def __init__(self, person, ticks):
        super(Wait, self).__init__(person)
    
        self._ticks_remaining = ticks
        
    def remaining(self):
        return self._ticks_remaining
    
    def act(self):
        self._ticks_remaining -= 1
        
        if self.remaining() <= 0:
            self.done()
            
        
class Step(Action):
    def __init__(self, person, vector):
        super(Step, self).__init__(person)
        
        self._vector        = vector
        self._progress      = 0.0
        self._tile_switched = False
        
    def vector(self):
        return self._vector

    def remaining_vector(self):
        return self.vector() * (Vector(-1, -1) + self.progress_vector())
    
    def done_vector(self):
        return self.vector() * self.progress_vector()
        
    def progress(self):
        return self._progress
    
    def tile_switched(self):
        return self._tile_switched
    
    def progress_vector(self):
        return Vector(self.progress(), self.progress())
    
    def displacement(self):
        if self.tile_switched():
            return self.remaining_vector()
        
        return self.done_vector()
    
    def act(self):
        if self.vector().length() <= 0.0:
            self.person().done_acting()
            return
        
        self._progress += self.person().speed() / self.vector().length()
        
        if self._progress >= 0.5 and not self.tile_switched():
            self.person().move_by(self.vector())
            self._tile_switched = True
        
        if self._progress >= 1.0:
            self.person().done_acting()
    
    
class Route(Plan):
    _open_routes = 0
    
    def __init__(self, person, target):
        super(Route, self).__init__(person)
        
        self._target  = target
        self._route   = []
        self._ready   = None
        self._open    = [(self.person().tile(), 0, 0)]
        self._closed  = set()
        self._parents = {}
        
        self.set_ready(False)
        
    @classmethod
    def open_routes(cls):
        return cls._open_routes
    
    def set_ready(self, ready):
        if ready == self.route_done():
            return
        
        self._ready = ready
        
        if ready:
            self.__class__._open_routes -= 1
        else:
            self.__class__._open_routes += 1
            
    def color(self):
        return Color(255, 255, 255)
    
    def target(self):
        return self._target
    
    def route(self):
        return list(self._route)
    
    def route_done(self):
        return self._ready
    
    def path(self):
        return self.route()
    
    def next_action(self):
        if not self.route_done():
            self.find_route(self.person().mind() * 2 / self.open_routes())
            
            return Wait(self.person(), 1)
        
        route = self.route()
        
        if not route:
            return None
        
        self._route = route[1:]
        
        return route[0]
    
    def find_route(self, steps):
        def better(x, y):
            (_, sx, ex) = x
            (_, sy, ey) = y

            return int(sx - sy + ex - ey)
        
        start  = self.person().tile()
        goal   = self.target()
                
        while self._open and steps > 0:
            steps -= 1
            
            self._open.sort(better)
            (current, score, _) = self._open[0]
            self._open = self._open[1:]
            self._closed.add(current)
            
            if current == goal:
                vectors = []
                
                while current != start:
                    vectors.append(current.vector())
                    current = self._parents[current]
                
                vectors.reverse()
                
                last = start.vector()
                
                for vector in vectors:
                    self._route.append(Step(self.person(), vector - last))
                    last = vector
                
                self.set_ready(True)
                
                return
            
            for tile in current.neighbors():
                if not tile.terrain().is_passable():
                    continue
                
                if tile in self._closed:
                    continue
                
                crazy      = min(1.0, 2.0 / self.person().mind())
                difficulty = (random.random() * 7 * crazy + 
                              tile.difficulty() * (1 - crazy))
                vector     = (goal.vector() - current.vector())
                distance   = (tile.vector() - current.vector()).length()
                new_score  = score + difficulty * distance
                estimate   = abs(vector.x()) + abs(vector.y())
                found      = None
                skip       = False
                
                for nr in range(len(self._open)):
                    (x, x_score, _) = self._open[nr]
                    
                    if x != tile:
                        continue
                    
                    if x_score <= new_score:
                        skip = True
                        continue
                    
                    found = nr
                    break
                
                if found is not None:
                    self._parents[tile] = current
                    self._open[found] = (tile, new_score, estimate)
                elif not skip:
                    self._parents[tile] = current
                    self._open.append((tile, new_score, estimate))
            
        if steps > 0:
            self.set_ready(True)
                
    def __str__(self):
        return 'Route to %s in %d steps' % (str(self.target()).lower(), 
                                            len(self.route()))


class Panic(Plan):
    def __init__(self, person):
        super(Panic, self).__init__(person)
        
        self._steps_left = max(5, 100 / (person.mind() + person.spirit()))
        
    def color(self):
        return Color(128, 0, 0)
    
    def steps_left(self):
        return self._steps_left
    
    def reduce_steps(self):
        self._steps_left -= 1
        
    def next_action(self):
        if self.steps_left() <= 0:
            return None
        
        tiles = list(self.person().tile().weighted_neighbors())
        
        if not tiles:
            return None
        
        self.reduce_steps()
        
        return Step(self.person(), random.choice(tiles).vector() -
                    self.person().tile().vector())
    
    def __str__(self):
        return 'Panic for %d steps' % self.steps_left()
    