'''Buildings in River.'''

from concepts.color import *
from concepts.geometry import *

from goods import *

from river.actions.jobs import *

from terrain import *

import random


class Building(object):
    _types = {'living': set(),
              'distribution': set(),
              'infra': set(),
              'industry': set()}
    
    def __init__(self, city, tile, jobs = []):
        self._city = city
        self._tile = tile
        self._jobs = jobs
        
    @classmethod
    def categories(cls):
        return ['infra', 'living', 'industry', 'distribution']
    
    @classmethod
    def types(cls, category):
        xs = list(cls._types[category])
        xs.sort()
        
        return xs
    
    @classmethod
    def add_type(cls, type, category):
        cls._types[category].add(type)
        
    @classmethod
    def can_be_built_on(cls, terrain):
        return (terrain.__class__ is cls.built_on())
    
    @classmethod
    def stands_free(cls):
        return True

    @classmethod
    def built_on(cls):
        return Grass
    
    @classmethod
    def size(self):
        return 1
    
    @classmethod
    def color(self):
        return Color(255, 0, 0)
    
    @classmethod
    def difficulty(self):
        return 1
    
    @classmethod
    def name(cls):
        return 'Building'
    
    @classmethod
    def cost(self):
        return [Wood(3)]
    
    def destroy(self):
        self.city().destroy(self)
        self._tile = None
        self._jobs = None
        
    def center(self):
        return self._tile.vector() + Vector(self.size() / 2.0,
                                            self.size() / 2.0)
    
    def city(self):
        return self._city
    
    def tiles(self):
        return self._tile.tiles_lower_right(self.size())

    def area(self):
        if self._tile is None:
            return None
        
        return Area(self._tile.vector(), Vector(self.size(), self.size()))
    
    def is_at(self, tile):
        return (tile in self.tiles())
    
    def jobs(self):
        return list(self._jobs)
    
    def housing_space(self):
        return 0
    
    def __str__(self):
        return '%s at %s' % (self.__class__.__name__, self.area())


class ConstructionSite(Building):
    def __init__(self, city, tile, type):
        super(ConstructionSite, self).__init__(city, tile,
                                               [Construction(self)])
    
        self._type           = type
        self._progress       = 0.0
        self._cost_remaining = self.type().cost()
        self._color          = self.type().color()
        self._total          = 0
        
        for x in self._cost_remaining:
            self._total += x.number()
        
    @classmethod
    def difficulty(cls):
        return 1
    
    @classmethod
    def name(cls):
        return 'Construction Site'

    @classmethod
    def built_on(cls):
        return None
        
    @classmethod
    def size(cls):
        return None

    def goods_needed(self):
        return list(self._cost_remaining)
    
    def total_needed(self):
        return self._total
    
    def number_needed(self):
        total = 0
        
        for x in self._cost_remaining:
            total += x.number()
        
        return total
    
    def type(self):
        return self._type
    
    def size(self):
        return self.type().size()
    
    def progress(self):
        return self._progress
    
    def needs_next_resource(self):
        max = 1 - float(self.number_needed()) / float(self.total_needed())
        
        return (self.progress() >= max)
    
    def needs(self, resource):
        for x in self.goods_needed():
            if x.__class__ is resource.__class__:
                return True
            
        return False
    
    def consume_resource(self):
        predicate = lambda x: x.place() in set(self.tiles()) and self.needs(x)
        resource  = self.jobs()[0].person().nearest_good(predicate)
        
        if resource is None:
            return
        
        resource.decrease(1)
        
        for x in self.goods_needed():
            if x.__class__ is resource.__class__:
                x.decrease(1)
                
                if x.number() < 1:
                    self._cost_remaining.remove(x)
                    
                return
            
    def add_progress(self, x):
        self._progress += x
        
        max = 1 - float(self.number_needed()) / float(self.total_needed())

        if self.progress() > max:
            self.consume_resource()
            
        if self.progress() > max:
            self._progress = max
            
        if self.progress() >= 1:
            self._progress = 1.0
            tile = self._tile
            type = self.type()
            
            self.city().log('%s constructed.' % (self.type().name()))
        
            self.color = lambda: type.color()            
            self.destroy()
            self.city().build(tile, type)
            
    def color(self):
        return Color(96, 96, 96).mix(self._color, 0.1 + self.progress() * 0.9)


class Construction(Job):
    def __init__(self, site):
        super(Construction, self).__init__(site)
    
        self._target = None
        
    def title(self):
        return 'Builder'
    
    def color(self):
        return Color(64, 64, 255)
    
    def target(self):
        if self._target is None:
            self._target = random.choice(list(self.building().tiles()))
            
        return self._target
    
    def office(self):
        return self.target()
    
    def work_done(self):
        mind = self.person().mind()
        body = self.person().body()
        
        self.building().add_progress(random.randint(body, mind + body) / 100.0)
    
    def work_speed(self):
        return 1.0 / self.building().size()

    def plan(self):
        if self.building().needs_next_resource():
            self.building().consume_resource()
            
        if self.building().needs_next_resource():
            needed   = self.building().goods_needed()[0]
            resource = self.person().claim_resource(needed.number(),
                                                    lambda x: x.__class__ is needed.__class__)
            
            if resource is None:
                return None
            
            return MoveResource(self.person(), resource, self, True)
        else:
            return super(Construction, self).plan()
        
    
class House(Building):
    def __init__(self, city, tile):
        super(House, self).__init__(city, tile)
    
        self._inhabitants = set()
        
    @classmethod
    def name(cls):
        return 'House'
    
    @classmethod
    def size(self):
        return 2
    
    @classmethod
    def color(self):
        return Color(128, 0, 0)
    
    @classmethod
    def cost(self):
        return [Wood(4)]
    
    def inhabitants(self):
        return set(self._inhabitants)
    
    def housing_space(self):
        return 2 - len(self.inhabitants())

    def move_in(self, person):
        self._inhabitants.add(person)
        
    def move_out(self, person):
        self._inhabitants.discard(person)


Building.add_type(House, 'living')
    

class Road(Building):
    def __init__(self, city, tile):
        super(Road, self).__init__(city, tile)
    
    @classmethod
    def name(cls):
        return 'Road'
    
    @classmethod
    def stands_free(cls):
        return False
    
    @classmethod
    def size(self):
        return 1
    
    @classmethod
    def cost(self):
        return [Wood(1)]
    
    @classmethod
    def color(self):
        return Color(128, 128, 128)
    
    @classmethod
    def difficulty(self):
        return 0.5


Building.add_type(Road, 'infra')
    

class Bridge(Building):
    def __init__(self, city, tile):
        super(Bridge, self).__init__(city, tile)
    
    @classmethod
    def name(cls):
        return 'Bridge'
    
    @classmethod
    def stands_free(cls):
        return False
    
    @classmethod
    def built_on(cls):
        return River
    
    @classmethod
    def size(self):
        return 1
    
    @classmethod
    def cost(self):
        return [Wood(1)]
    
    @classmethod
    def color(self):
        return Color(72, 64, 0)
    
    @classmethod
    def difficulty(self):
        return 0.6


Building.add_type(Bridge, 'infra')
    