'''The city in River.'''

from building import *
from map import *
from people import *

from actions.jobs import *


class City(object):
    def __init__(self, name, map, time):
        self._name      = name
        self._map       = map
        self._time      = time
        self._buildings = set()
        self._persons   = set()
        self._goods     = set()
        self._jobs      = JobBoard(self)
        self._logger    = None
        
        map.set_city(self)
    
    def logger(self):
        return self._logger
    
    def set_logger(self, logger):
        self._logger = logger
        
    def has_logger(self):
        return (self._logger is not None)
    
    def log(self, string, color = None):
        if self.has_logger():
            self.logger().log(string, color)
            
    def name(self):
        return self._name
    
    def map(self):
        return self._map
    
    def time(self):
        return self._time
    
    def jobs(self):
        return self._jobs
    
    def buildings(self):
        return set(self._buildings)
    
    def building_on(self, tile):
        result = filter(lambda x: x.is_at(tile), self.buildings())
        
        if not result:
            return None
        
        return result[0]
    
    def persons(self):
        return set(self._persons)
    
    def persons_on(self, tile):
        return filter(lambda x: x.is_at(tile), self.persons())        
        
    def persons_with(self, predicate = lambda x: True):
        return filter(predicate, self.persons())
    
    def persons_of_type(self, type):
        return self.persons_with(lambda x: x.job_description() == type)
    
    def goods(self):
        return set(self._goods)
    
    def goods_at(self, place):
        return filter(lambda x: x.is_at(place), self.goods())
    
    def can_build(self, tile, building_type):
        tiles = set(tile.tiles_lower_right(building_type.size()))
        built = any(map(lambda x: x.is_built(), tiles))
        terrain_matches = all(map(lambda x: building_type.can_be_built_on(x.terrain()), tiles))
                                  
        return (not built and terrain_matches)
    
    def build(self, tile, building_type):
        if not self.can_build(tile, building_type):
            return
        
        building = building_type(self, tile)
        
        self._buildings.add(building)
        
        for tile in building.tiles():
            tile.set_building(building)
        
        for job in building.jobs():
            self.jobs().add_job(job)
    
    def start_construction(self, tile, building_type):
        if not self.can_build(tile, building_type):
            return
        
        site = ConstructionSite(self, tile, building_type)
        
        self._buildings.add(site)
        
        for tile in site.tiles():
            tile.set_building(site)
        
        for job in site.jobs():
            self.jobs().add_job(job)
    
    def destroy(self, building):
        self._buildings.discard(building)
        
        for tile in building.tiles():
            tile.set_building(None)
        
        for job in building.jobs():
            self.jobs().cancel_job(job)
        
    def give_birth(self, tile, name = None):
        self._persons.add(Person(name, self, tile))
    
    def create(self, tile, good):
        good.move_to(tile.place())
        self._goods.add(good)        
        self.consolidate_goods_at(tile)
        
    def consolidate_goods_at(self, tile):
        existing = self.goods_at(tile)
        found    = None
        
        for x in existing:
            if x.has_owner():
                continue
            
            if found is None:
                found = x
                continue
            
            if found.__class__ is not x.__class__:
                continue
            
            found.increase(x.number())
            x.destroy()
        
    def consume(self, good):
        good.move_to(None)
        self._goods.remove(good)
        
    def nearest_usable_good(self, person, predicate = lambda x: True):
        usable = filter(lambda x: predicate(x) and x.can_be_used_by(person),
                        self.goods())
        
        if not usable:
            return None
        
        usable.sort(lambda x, y: int(person.tile().distance(x.place()) - 
                    person.tile().distance(y.place())))
        
        return usable[0]
        
    def buildings_with(self, predicate = lambda x: True):
        return filter(predicate, self.buildings())
    
    def buildings_of_type(self, type):
        return self.buildings_with(lambda x: x.__class__ is type)
    
    def nearest_building(self, person, predicate = lambda x: True):
        buildings = filter(predicate, self.buildings())
        
        if not buildings:
            return None
        
        buildings.sort(lambda x, y: int(person.tile().distance(x.center()) - 
                                        person.tile().distance(y.center())))
        
        return buildings[0]
    
    def seek_home(self, person):
        home = self.nearest_building(person, lambda x: x.housing_space() > 0)
        
        if home is None:
            return
        
        home.move_in(person)
        person.set_home(home)
        