'''River UI parts.'''

from concepts.geometry import *
from concepts.color import *

from ui.components import *
from ui.texture import *
from ui.window import *

from building import *
from map import *
from views import *


class RiverWindow(PrettyWindow):
    def __init__(self, size, scenario, logger):
        frame_color      = Color(255, 255, 0)
        background_color = Color(64, 64, 216)
        border_size      = 7
        
        super(RiverWindow, self).__init__('Twisted River Community', size,
                                          frame_color, background_color)
        
        background_color = background_color.dimmed()
        
        client_area = self.full_area().cropped(border_size)
        left        = client_area.cropped(0, 0, 250 + border_size, 0)
        right       = client_area.with_width(250).moved(left.extent().x() + 
                                                        border_size, 0)
        view        = ViewPanel(right.cropped(0,  32 + border_size, 
                                              0, 250 + border_size),
                                frame_color, background_color)
        main        = MainPanel(left.cropped(0,  32 + border_size, 
                                             0, 100 + border_size),
                                frame_color, background_color, scenario,
                                view.on_select)
        resources   = ResourcesPanel(left.with_height(32), frame_color,
                                     background_color)
        text        = TextPanel(left.with_height(100).moved(0, 32 + 
                                                            2 * border_size + 
                                                            main.area().extent().y()), 
                                frame_color, background_color, logger)
        button      = ButtonPanel(right.with_height(32), frame_color,
                                  background_color)
        building    = BuildingPanel(right.with_height(250).moved(0, 32 + 
                                                                 2 * border_size + 
                                                                 view.area().extent().y()), 
                                    scenario, main)
        
        self.add(main)
        self.add(resources)
        self.add(text)
        self.add(button)
        self.add(view)
        self.add(building)


class RiverButton(Button):
    def __init__(self, area):
        super(RiverButton, self).__init__(area, Color(255, 255, 255),
                                          Color(64, 64, 216))


class RiverPanel(Panel):
    def __init__(self, area, scenario):
        super(RiverPanel, self).__init__(area, Color(255, 255, 0),
                                          Color(64, 64, 216).dimmed())
        
        self._scenario = scenario
    
    def scenario(self):
        return self._scenario
    
    def city(self):
        return self.scenario().city()
    
    def tile(self, x, y):
        return self.scenario().map().tile(x, y)
    
    def do_event(self, event):
        self.scenario().time().add_ui_event(event)


class MainPanel(Panel):
    def __init__(self, area, frame_color, background_color, scenario, 
                 on_select):
        super(MainPanel, self).__init__(area, frame_color, background_color)
        
        self._view = MapView(self.client_area(), scenario.city(),
                             15, Color(255, 255, 255), on_select)
        
        self.add(self.view())

    def view(self):
        return self._view
    
    def set_build_callback(self, callback, building):
        self.view().set_build_callback(callback, building)


class ResourcesPanel(Panel):
    def __init__(self, area, frame_color, background_color):
        super(ResourcesPanel, self).__init__(area, frame_color,
                                             background_color)


class ButtonPanel(Panel):
    def __init__(self, area, frame_color, background_color):
        super(ButtonPanel, self).__init__(area, frame_color,
                                          background_color)


class ViewPanel(Panel):
    def __init__(self, area, frame_color, background_color):
        super(ViewPanel, self).__init__(area, frame_color,
                                        background_color)
        self._view = None
        
    def show_view(self, view):
        if self._view is not None:
            self.remove(self._view)
            self._view.dispose()
            
        self._view = view
        
        if view is not None:
            self.add(view)
        
    def on_select(self, thing):
        if thing is None:
            self.show_view(None)
        else:
            self.show_view(PersonView(self.client_area(), thing))
    
        
class TextPanel(Panel):
    def __init__(self, area, frame_color, background_color, logger):
        super(TextPanel, self).__init__(area, frame_color,
                                        background_color)
        
        self.add(TextView(self.client_area(), logger))
            

class ButtonRow(object):
    def __init__(self, panel, nr):
        self._buttons = []
        self._panel   = panel
        self._nr      = nr
        
    def buttons(self):
        return self._buttons
    
    def panel(self):
        return self._panel
    
    def nr(self):
        return self._nr
    
    def add_button(self, building):
        button_area = self.panel().client_area().cropped(7).with_width(56).with_height(56)
        button_area = button_area.moved((button_area.extent().x() + 3) * len(self.buttons()),
                                        (button_area.extent().y() + 3) * self.nr())
        
        button = BuildingButton(button_area, self.panel(), building)
        
        self.panel().add_button(button)
        self._buttons.append(button)
    
    
class BuildingPanel(RiverPanel):
    def __init__(self, area, scenario, main):
        super(BuildingPanel, self).__init__(area, scenario)
        
        self._buttons  = set()
        self._rows     = {}
        self._building = None
        self._main     = main
        
        nr = 0
        
        for category in Building.categories():
            self._rows[category] = ButtonRow(self, nr)
            
            nr += 1
            
            for type in Building.types(category):
                self._rows[category].add_button(type)
                        
    def main(self):
        return self._main
        
    def buttons(self):
        return set(self._buttons)
    
    def add_button(self, button):
        self._buttons.add(button)
        self.add(button)
        
    def set_building(self, building):
        if building == self.building():
            building = None
            
        self._building = building
        
        for button in self.buttons():
            button.highlight_if_building(building)
            
        if building is not None:
            self.city().log('%s costs: %s' % (building.name(),
                                             ', '.join(map(str, building.cost()))))
            
            self.main().set_build_callback(self.build, building)
        else:
            self.main().set_build_callback(None, None)
            
    def build(self, tile, building):
        if tile is None:
            self.set_building(None)
            return
        
        self.do_event(lambda: self.city().start_construction(tile, building))

    def building(self):
        return self._building
    
            
class BuildingButton(RiverButton):
    def __init__(self, area, panel, building):
        super(BuildingButton, self).__init__(area)
        
        self._panel    = panel
        self._building = building
    
    def panel(self):
        return self._panel
    
    def building(self):
        return self._building
    
    def highlight_if_building(self, building):
        self.set_state(building == self.building())
        
    def do_click(self, vector):
        self.panel().set_building(self.building())
        
    def render_contents(self, renderer):
        area = self.client_area().cropped(11, 4, 11, 18)
        
        renderer.draw_filled(area, self.building().color())
        renderer.draw_rectangle(area, Color(0, 0, 0))
        renderer.draw_text(self.building().name(), Vector(4, 38),
                           Color(255, 255, 255),
                           self.client_area().extent().x() - 4, True, True)
    
    def render(self):
        if not self.panel().scenario().is_allowed(self.building()):
            return
        
        super(BuildingButton, self).render()
        
    