'''Basic physical world of River.'''

from itertools import *

from concepts.color import *
from concepts.geometry import *

from terrain import *


class CityMap(object):
    def __init__(self, extent):
        self._city   = None
        self._extent = extent
        self._tiles  = [[Tile(self, Vector(x, y))
                         for y in range(extent.y())]
                         for x in range(extent.x())]
        
    def city(self):
        return self._city
    
    def set_city(self, city):
        if self._city is None:
            self._city = city
    
    def extent(self):
        return self._extent
    
    def place(self, x, y):
        return MovablePlace(self, Vector(x, y))
    
    def tile(self, x, y):
        if x < 0 or y < 0 or x >= self.extent().x() or y >= self.extent().y():
            return Tile(self, Vector(x, y), Outside)
        
        return self._tiles[x][y]

    def __iter__(self):
        return chain(*self._tiles)
    
    
class Place(object):
    def __init__(self, map, origin):
        self._map    = map
        self._origin = origin
    
    def map(self):
        return self._map
    
    def city(self):
        return self.map().city()
    
    def vector(self):
        return self._origin
        
    def x(self):
        return self.vector().x()
    
    def y(self):
        return self.vector().y()
    
    def distance(self, other):
        return (self.vector() - other.vector()).length()
    
    def place(self):
        return self
    
    def __eq__(self, other):
        return (other is not None and
                self.city() == other.city() and
                self.x() == other.x() and
                self.y() == other.y())
        
    def __ne__(self, other):
        return not (self == other)
    
    def __hash__(self):
        return self.x() + self.y()

    
class Tile(Place):
    def __init__(self, map, origin, terrain = Grass):
        super(Tile, self).__init__(map, origin)
        
        self._terrain  = terrain()
        self._building = None
    
    def terrain(self):
        return self._terrain
    
    def become(self, terrain):
        self._terrain = terrain()
    
    def build(self, building_type):
        self.city().build(self, building_type)
        
    def exists(self):
        return (self._terrain.__class__ != Outside)
    
    def place(self):
        return self.map().place(self.x(), self.y())
    
    def neighbors(self):
        for x in [-1, 0, 1]:
            for y in [-1, 0, 1]:
                yield self.in_direction(Vector(x, y))
            
    def weighted_neighbors(self):
        for tile in self.neighbors():
            if not tile.terrain().is_passable():
                continue
        
            num = int(5 / tile.difficulty())
            
            for _ in range(num):
                yield tile
        
    def in_direction(self, vector):
        return self.map().tile(self.x() + vector.x(),
                               self.y() + vector.y())

    def spawn(self, resource):
        self.city().create(self, resource)
        
    def is_built(self):
        return (self.building() is not None)
    
    def building(self):
        return self._building
    
    def set_building(self, building):
        self._building = building
    
    def tiles_lower_right(self, size):
        for x in range(size):
            for y in range(size):
                yield self.in_direction(Vector(x, y))
    
    def difficulty(self):
        if self.is_built():
            return self.building().difficulty()
        
        return self.terrain().difficulty()
    
    def __repr__(self):
        return 'Tile' + str(self.vector())
    
    def __str__(self):
        return '%s tile at %s' % (self.terrain(), self.vector())


class MovablePlace(Place):
    def __init__(self, map, origin):
        super(MovablePlace, self).__init__(map, origin)
        
    def tile(self):
        return self.map().tile(self.x(), self.y())
    
    def move_by(self, vector):
        self._origin += vector
        
    def move_to(self, vector):
        self._origin = vector
        