'''Huggable people in river. :-)'''

from concepts.color import *
from concepts.geometry import *

from actions.consume import *
from actions.move import *
from attributes import *
from names import *

import random
import math


subject = object


class Person(subject):
    def __init__(self, name, city, tile):
        if name is None:
            name = random.choice(NAMES)
            
        self._name       = name
        self._city       = city
        self._tile       = tile
        self._action     = None
        self._plan       = None
        self._home       = None
        self._job        = None
        self._nutrition  = Nutrition(self, 0.8)
        self._energy     = Energy(self, 1.0)
        self._social     = Social(self, 0.5)
        self._attributes = Attributes()
        self._start_tile = tile
        self._backpack   = None
        
        self.time().do_every(1, self.act)
    
    def name(self):
        return self._name
    
    def color(self):
        if self.has_job():
            return self.job().color()
        
        return Color(255, 255, 255)
    
    def base_speed(self):
        return (  0.050
                + 0.003 * self.body()
                - 0.002 * abs(self.nutrition().amount() - 0.5) * 10
                + 0.004 * min(0, self.energy().amount() - 0.5) * 10)
    
    def backpack(self):
        return self._backpack
    
    def speed(self):
        return self.base_speed() / self.tile().difficulty()
    
    def city(self):
        return self._city
    
    def time(self):
        return self.city().time()
    
    def attributes(self):
        return self._attributes
    
    def nutrition(self):
        return self._nutrition
    
    def energy(self):
        return self._energy
    
    def social(self):
        return self._social
    
    def body(self):
        return self.attributes().body()
    
    def mind(self):
        return self.attributes().mind()
    
    def spirit(self):
        return self.attributes().spirit()
    
    def job(self):
        return self._job
    
    def has_job(self):
        return (self.job() is not None)
    
    def seek_employment(self):
        self.city().jobs().seek_employment(self)
    
    def seek_home(self):
        self.city().seek_home(self)
        
    def employed(self, job):
        self._job = job
        
    def fired(self, job):
        self._job = None
    
    def job_description(self):
        if self.job() is None:
            return 'Villager'
        
        return self.job().title()
    
    def home(self):
        return self._home
    
    def set_home(self, home):
        self._home = home
    
    def has_home(self):
        return (self._home is not None)
    
    def nearest_good(self, predicate = lambda x: True):
        return self.city().nearest_usable_good(self, predicate)
    
    def consume(self, good):
        if good is None:
            return
        
        self.set_plan(GoAndConsume(self, good))
        
    def tile(self):
        return self._tile
    
    def set_tile(self, tile):
        self._tile = tile
        
        if self.backpack() is not None:
            self.backpack().move_to(tile.place())
        
    def start_tile(self):
        return self._start_tile
    
    def move_by(self, vector):
        vector += self.tile().vector()
        
        self.set_tile(self.city().map().tile(vector.x(), vector.y()))
    
    def take(self, resource):
        self._backpack = resource
        
    def drop(self, resource, keep = False):
        self._backpack = None
        
        if not keep:
            resource.set_owner(None)
            
        self.city().consolidate_goods_at(self.tile())
        
    def is_at(self, tile):
        return (self.tile() == tile)
    
    def action(self):
        return self._action
    
    def has_action(self):
        return (self.action() is not None)
    
    def decide_plan(self):
        if self.nutrition().is_hungry():
            self.go_eating()
            return
        
        if not self.has_home():
            self.seek_home()
            
        if not self.has_job():
            self.seek_employment()
            
        if self.has_job() and self.job().has_plan():
            self.go_working()
            return
        
        if self.has_home():
            self.go_home()
            return
        
        self.go_wandering()
            
    def claim_resource(self, number, predicate):
        resource = self.nearest_good(predicate)
        
        if resource is None:
            return None
         
        return resource.take(number, self)
        
    def go_eating(self):
        food = self.claim_resource(5, lambda x: x.is_edible())
        
        if food is None:
            self.set_plan(Panic(self))
            return
        
        self.consume(food)
    
    def go_working(self):
        self.set_plan(self.job().plan())
    
    def go_wandering(self):
        if random.randint(1, 10) == 1:
            buildings = self.city().buildings()
            
            if not buildings:
                return
            
            building = random.choice(list(buildings))
            tile     = random.choice(list(building.tiles()))
        else:
            tile = random.choice(filter(lambda x: x.exists(),
                                        self.tile().weighted_neighbors()))
            
        self.set_plan(Route(self, tile))
    
    def go_home(self):
        tile = random.choice(list(self.home().tiles()))
        
        self.set_plan(Route(self, tile))

    def act(self):
        self.nutrition().reduce(0.0001)
        self.energy().reduce(0.0001)
        self.social().reduce(0.0001)
        self.social().spend_time_with(self.city().persons_on(self.tile()))

        if not self.has_action():
            if not self.has_plan():
                self.decide_plan()

            if not self.has_plan():
                return
            
            if not self.has_action():
                self.done_acting()
                return
            
            self._action = self.plan().next_action()
        
        action = self.action()
        
        if action is None:
            return
        
        self.nutrition().reduce(0.0002)
        self.energy().reduce(0.00003)
        action.act()
        
    def done_acting(self):
        if self.has_plan():
            self._action = self.plan().next_action()
            
            if not self.has_action():
                self.set_plan(None)
        else:
            self._action = None
            
        self._start_tile = self.tile()
        
    def plan(self):
        return self._plan
    
    def has_plan(self):
        return (self.plan() is not None)
    
    def set_plan(self, plan):
        self._plan = plan
        
    def displacement(self):
        if not self.has_action():
            return Vector(0, 0)
        
        return self.action().displacement()

    def __str__(self):
        return ('%s (carrying: %s; job: %s; plan: %s; action: %s)' % 
                (self.name(), self.backpack(), self.job(),
                 self.plan(), self.action()))


def _above(n):
    return lambda self: self.amount() > n

def _below(n):
    return lambda self: self.amount() < n

def _between(n, m):
    return lambda self: n < self.amount() < m


class Stat(object):
    def __init__(self, person, amount):
        self._person = person
        self._amount = amount

    def person(self):
        return self._person
    
    def amount(self):
        return self._amount
    
    def add(self, amount):
        self._amount = max(min(self.amount() + amount, 1), 0)
    
    def reduce(self, amount):
        self._amount = max(min(self.amount() - amount, 1), 0)

    def color(self):
        if self.amount() > 0.75:
            return Color(0, 128, 0)
        
        if self.amount() > 0.50:
            return Color(0, 216, 0)
        
        if self.amount() < 0.10:
            return Color(128, 0, 0)
            
        if self.amount() < 0.25:
            return Color(255, 128, 0)
        
        return Color(255, 255, 0)

    
class Nutrition(Stat):
    is_full     = _above(0.75)
    is_satiated = _above(0.50)
    is_hungry   = _below(0.25)
    is_starving = _below(0.10)
    
    def consumption_needed(self, good):
        return int((1 - self.amount()) / good.nutrional_value())
        
        
class Energy(Stat):
    is_tired = _below(0.25)


class Social(Stat):
    def spend_time_with(self, people):
        value = 0
        
        for x in people:
            if self.person() == x:
                continue
            
            if random.randint(0, x.spirit()) < 2:
                value -= 0.0003
            
            delta  = max(0, x.spirit() - max(0, self.amount() - 0.25) * 20)
            value += delta * 0.00025
                
        value /= math.sqrt(len(people))
        
        self.add(value)
        