'''UI views on objects.'''

from concepts.geometry import *
from concepts.color import *

from ui.components import *
from ui.texture import *
from ui.window import *

import pygame.font
import math
import random


class PersonView(Component):
    def __init__(self, area, person):
        super(PersonView, self).__init__(area)

        self._person = person
    
    def person(self):
        return self._person

    def draw_attribute(self, text, y, stat):
        size = self.font_size()

        self.draw_text(text + ':', Vector(7, y), Color(255, 255, 255),
                       self.area().extent().x() - 89)
        
            
        self.draw_bar(stat.amount(), Vector(75, y),
                      Vector(self.area().extent().x() - 82, size), stat.color(),
                      Color(255, 255, 255))
        
    def render(self):
        pos  = 7
        size = self.font_size()
        
        self.draw_circle(self.person().color(), self.person().color().dimmed(),
                         Vector(self.area().extent().x() / 2, pos + size),
                         size, 2)
        
        pos += 7 + 2 * size
        
        self.draw_text(self.person().name(), Vector(7, pos), 
                       Color(255, 255, 255),
                       self.area().extent().x() - 7, True)
        
        pos += self.font_size() * 1.3
        
        if self.person().has_job():
            title = self.person().job().title()
        else:
            title = 'Villager'
            
        self.draw_text(title, Vector(7, pos), self.person().color(),
                       self.area().extent().x() - 7, True)
        
        pos += self.font_size() * 1.3

        self.draw_text(str(self.person().tile().vector()), Vector(7, pos), 
                       Color(255, 255, 255).dimmed(),
                       self.area().extent().x() - 7, True)
        
        pos += self.font_size() * 1.3 + 12

        self.draw_text('Body: %d    Mind: %d    Spirit: %d' %
                       (self.person().body(), self.person().mind(), 
                        self.person().spirit()),
                       Vector(7, pos), Color(255, 255, 255),
                       self.area().extent().x() - 7, True)
        
        pos += self.font_size() * 1.3 + 12
        
        self.draw_attribute('Food', pos, self.person().nutrition())
        
        pos += self.font_size() * 1.3
        
        self.draw_attribute('Energy', pos, self.person().energy())
        
        pos += self.font_size() * 1.3
        
        self.draw_attribute('Social', pos, self.person().social())
        
        pos += self.font_size() * 1.3
    
    
class TextView(Component):
    def __init__(self, area, logger):
        super(TextView, self).__init__(area)

        self._logger = logger
        
    def logger(self):
        return self._logger
    
    def render(self):
        pos  = 5
        last = self.logger().last(5)
        
        for line, color in last:
            self.draw_text(line, Vector(5, pos), color.brightened(),
                           self.area().extent().x() - 10)
            pos += self.font_size() * 1.3


class MapView(Component):
    def __init__(self, area, city, tile_size, border_color, on_select):
        super(MapView, self).__init__(area)
        
        self._city           = city
        self._tile_size      = tile_size
        self._border_color   = border_color
        self._textures       = {}
        self._display_vector = None
        self._on_select      = on_select
        self._selection      = None
        self._callback       = None
        self._building       = None
        
    def callback(self):
        return self._callback
    
    def building(self):
        return self._building
    
    def select(self, thing):
        self._selection = thing
        self._on_select(thing)
    
    def selection(self):
        return self._selection
    
    def is_selected(self, thing):
        return (thing == self.selection())
    
    def border_color_of(self, thing):
        if self.is_selected(thing):
            return Color(0, 0, 0)
        
        return thing.color().dimmed()
    
    def log(self, string, color = None):
        self.city().log(string, color)
        
    def city(self):
        return self._city
    
    def city_map(self):
        return self.city().map()
    
    def tile_size(self):
        return self._tile_size
    
    def tile_vector(self, size = 1):
        return Vector(self.tile_size() * size, self.tile_size() * size)
    
    def tile_origin(self, tile):
        return self.display_vector() + Vector(tile.x() * self.tile_size(),
                                              tile.y() * self.tile_size())
        
    def border_color(self):
        return self._border_color
    
    def smaller_area(self):
        return (self.display_area() < self.area())
    
    def texture(self, color, size = 1, growth = 0):
        if not self._textures.has_key(color):
            self._textures[color] = Texture(self.tile_vector(size) +
                                            Vector(growth, growth), color)
            
        return self._textures[color]
    
    def display_vector(self):
        if self._display_vector is None:
            x  = min(self.area().extent().x(),
                     self.city_map().extent().x() * self.tile_size())
            y  = min(self.area().extent().y(),
                     self.city_map().extent().y() * self.tile_size())
            dx = self.area().extent().x() - x
            dy = self.area().extent().y() - y
        
            self._display_vector = Vector(dx / 2.0, dy / 2.0)
        
        return self._display_vector
    
    def display_area(self):
        vector = self.display_vector()
        
        return self.area().cropped(vector.x(), vector.y())
    
    def tile_area(self, tile, size = 1):
        return Area(self.tile_origin(tile) + self.origin(),
                    self.tile_vector(size))
    
    def tile_at(self, vector):
        tile_vector = (vector - self.display_vector()) / self.tile_size()
        x, y        = tile_vector.int_tuple()
        tile        = self.city_map().tile(x, y)
        
        if not tile.exists():
            return None
        
        return tile
    
    def mouse_tile(self):
        vector = self.mouse()
        
        if vector is None:
            return None
        
        return self.tile_at(vector)
    
    def render_tiles(self):
        for tile in self.city_map():
            self.blit(self.texture(tile.terrain().color()),
                      self.tile_origin(tile))
        
    def render_buildings(self):
        for building in self.city().buildings():
            area = building.area()
            
            if area is None:
                continue
            
            tile = area.origin()
            
            if building.stands_free():
                self.blit(self.texture(building.color(), building.size(), -2),
                          self.tile_origin(tile) + Vector(1, 1))
            else:
                self.blit(self.texture(building.color(), building.size()),
                          self.tile_origin(tile))
                
    def render_goods(self):
        shift = self.tile_vector() * Vector(0.5, 0.5)

        for good in self.city().goods():
            place  = good.place()
            number = good.number()
            
            if place is None:
                continue
            
            if number <= 0:
                continue

            self.draw_lozengy(good.color(), good.color().dimmed(),
                              self.tile_origin(place) + shift,
                              min(math.log(number, 2) + 2, 
                                  self.tile_size() / 2 - 2))
        
    def render_paths(self):
        shift = self.tile_vector() * Vector(0.5, 0.5)

        for person in self.city().persons():
            if not self.is_selected(person):
                continue
            
            action   = person.action()
            plan     = person.plan()
            tile     = person.start_tile()
            last     = self.tile_origin(tile)
            path     = []
            color    = Color(0, 0, 0)
            dimmed   = color.dimmed()
            
            if action is not None:
                last += self.tile_vector() * action.done_vector()
                path += [self.tile_vector() * -action.remaining_vector()]
                
            if plan is not None:
                color  = plan.color()
                dimmed = color.dimmed()
                path  += map(lambda x: self.tile_vector() * x.vector(),
                             plan.path())
            
            for element in path:
                current = last + element
                
                self.draw_line(dimmed, last + shift, current + shift)
                self.draw_circle(color, dimmed, last + shift, 3)
                self.draw_circle(color, dimmed, current + shift, 3)
                
                last = current
                
    def render_people(self):
        shift = self.tile_vector() * Vector(0.5, 0.5)

        for person in self.city().persons():
            tile     = person.tile()
            position = self.tile_vector() * person.displacement()
            
            self.draw_circle(person.color(), self.border_color_of(person),
                             self.tile_origin(tile) + shift + position,
                             self.tile_size() / 2)

    def render_construction(self):
        if self.callback() is None:
            return
        
        if not self.has_mouse():
            return
        
        building = self.building()
        tile     = self.mouse_tile()
        
        if not tile:
            return
        
        area = self.tile_area(tile, building.size())
        
        self.draw_rectangle(area, building.color())
        self.draw_rectangle(area.cropped(1), building.color())
        
        if not self.city().can_build(tile, building):
            cross = area.cropped(2).moved(-self.origin().x(),
                                          -self.origin().y())
            
            self.draw_line(Color(0, 0, 0), cross.origin(), cross.far_point() + Vector(0, 1))
            self.draw_line(Color(0, 0, 0), Vector(cross.origin().x(),
                                                  cross.far_point().y() + 1),
                           Vector(cross.far_point().x(),
                                  cross.origin().y()))

    def do_right_click(self, vector):
        if self.callback() is not None:
            self.callback()(None, self.building())
        else:
            self.select(None)        
        
    def do_left_click(self, vector):
        tile = self.tile_at(vector)
        
        if not tile or not tile.exists():
            return
        
        if self.callback() is not None:
            self.callback()(tile, self.building())
            return
        
        persons  = self.city().persons_on(tile)
        
        if persons:
            self.select(random.choice(list(persons)))
        else:
            self.select(None)
            
    def do_middle_click(self, vector):
        tile = self.tile_at(vector)
        
        if not tile or not tile.exists():
            return

        building = self.city().building_on(tile)
        persons  = self.city().persons_on(tile)
        goods    = self.city().goods_at(tile.place())
            
        self.log('Debug Information:')
        self.log('  ' + str(tile), tile.terrain().color())
        
        if building is None:
            self.log('  Nothing built here')
        else:
            self.log('  ' + str(building), building.color())
        
        if persons:
            self.log('  ' + ', '.join(map(str, persons)))
        else:
            self.log('  No people here')
            
        if goods:
            self.log('  ' + ', '.join(map(str, goods)))
        else:
            self.log('  No goods here')
            
    def set_build_callback(self, callback, building):
        self._callback = callback
        self._building = building
        
    def render(self):
        if self.smaller_area():
            self.draw_rectangle(self.display_area().cropped(-1),
                                self.border_color())
        
        self.render_tiles()
        self.render_buildings()
        self.render_goods()
        self.render_paths()
        self.render_people()
        self.render_construction()
        